;; Generated package description from validate.el  -*- no-byte-compile: t -*-
(define-package "validate" "1.0.4.0.2017030620828" "Schema validation for Emacs-lisp" '((emacs "24.1") (cl-lib "0.5") (seq "2.16")) :url "https://elpa.gnu.org/packages/validate.html" :keywords '("lisp") :authors '(("Artur Malabarba" . "emacs@endlessparentheses.com")) :maintainer '("Artur Malabarba" . "emacs@endlessparentheses.com"))
